/*!
 * Jodit Editor (https://xdsoft.net/jodit/)
 * License https://xdsoft.net/jodit/license.html
 * Copyright 2013-2018 Valeriy Chupurnov https://xdsoft.net
 */

import {Jodit} from '../Jodit';
import {Config} from '../Config';
import {debounce, dom, offset} from "../modules/Helpers";
import {ToolbarIcon} from "../modules/ToolbarCollection";
import * as consts from '../constants';
import {Dom} from "../modules/Dom";

declare module "../Config" {
    interface Config {
        addNewLine: boolean;
        addNewLineTagsTriggers: string[];
    }
}

/**
 * Create helper
 * @type {boolean}
 */
Config.prototype.addNewLine = true;

/**
 * Whar kind of tags it will be impact
 * @type {string[]}
 */
Config.prototype.addNewLineTagsTriggers = ['table','iframe', 'img', 'hr', 'jodit'];

/**
 * Create helper for adding new paragraph(Jodit.defaultOptions.enter tag) before iframe, table or image
 *
 * @param {Jodit} editor
 */

export function addNewLine(editor: Jodit) {
    if (!editor.options.addNewLine) {
        return;
    }

    const line: HTMLDivElement = <HTMLDivElement>dom('<div role="button" tabIndex="-1" title="' + editor.i18n("Break") + '" class="jodit-add-new-line"><span>' + ToolbarIcon.getIcon('enter') + '</span></div>', editor.ownerDocument);
    const span: HTMLSpanElement = <HTMLSpanElement>line.querySelector('span');
    const delta = 10;
    const isMatchedTag = new RegExp('^(' + editor.options.addNewLineTagsTriggers.join('|') + ')$', 'i');

    let timeout: number;
    let hidden: boolean = false;
    let preview: boolean = false;
    let current: HTMLElement|false;

    const show = () => {
        if (editor.options.readonly) {
            return;
        }

        if (editor.container.classList.contains('jodit_popup_active')) {
            return;
        }

        clearTimeout(timeout);
        line.classList.toggle('jodit-add-new-line_after', !preview);
        line.style.display = 'block';
        line.style.width = editor.editor.clientWidth + 'px';
        hidden = false;
    };

    const hideForce = () => {
        clearTimeout(timeout);
        line.style.display = 'none';
        hidden = true;
    };

    const canGetFocus = (elm: Node | null): boolean => {
        return elm !== null && Dom.isBlock(elm) && !/^(img|table|iframe|hr)$/i.test(elm.nodeName);
    };

    const hide = () => {
        if (hidden) {
            return;
        }
        clearTimeout(timeout);
        timeout = window.setTimeout(hideForce, 500)
    };


    editor.events
        .on('beforeDestruct', () => {
            line.parentNode && line.parentNode.removeChild(line);
        })
        .on('afterInit', () => {
            editor.container.appendChild(line);
            editor.events
                .on(line, 'mousemove', (e: MouseEvent) => {
                    e.stopPropagation();
                })
                .on(span, 'mousedown touchstart', (e: MouseEvent) => {
                    const p: HTMLElement = editor.editorDocument.createElement(editor.options.enter),
                        helper_node: Node = editor.editorDocument.createTextNode(consts.INVISIBLE_SPACE);

                    p.appendChild(helper_node);

                    if (preview && current && current.parentNode) {
                        current.parentNode.insertBefore(p, current);
                    } else {
                        editor.editor.appendChild(p)
                    }

                    editor.selection.setCursorIn(p);

                    editor.setEditorValue();
                    hideForce();
                    e.preventDefault();
                })
        })
        .on('change afterInit afterSetMode', debounce(() => {
            editor.events
                .on(editor.editor, 'scroll', () => {
                    hideForce();
                })
                .on(editor.container, 'mouseleave', () => {
                    hide();
                })
                .on(editor.editor, 'mousemove', debounce(function (this: HTMLElement, e: MouseEvent) {
                    let currentElement: HTMLElement = <HTMLElement>editor.editorDocument.elementFromPoint(e.pageX - editor.editorWindow.pageXOffset, e.pageY - editor.editorWindow.pageYOffset);

                    if (!currentElement || currentElement === line || !Dom.isOrContains(editor.editor, currentElement)) {
                        return;
                    }

                    current = currentElement;

                    if (!current || !current.nodeName.match(isMatchedTag) || !Dom.isOrContains(editor.editor, current)) {
                        current = <HTMLElement>Dom.closest(current, isMatchedTag, editor.editor);
                        if (!current) {
                            hide();
                            return;
                        }
                    }

                    if (current.nodeName.match(isMatchedTag)) {
                        const parentBox: Node | false = Dom.up(current, Dom.isBlock, editor.editor);
                        if (parentBox && parentBox !== editor.editor) {
                            current = <HTMLElement>parentBox;
                        }
                    }


                    // const parentPosition = offset(<HTMLElement>editor.workplace, editor);
                    const position: Bound = offset(<HTMLElement>current, editor, editor.editorDocument);
                    let top: false|number = false;

                    if (Math.abs(e.pageY - position.top) < delta) {
                        top = position.top;
                        preview = true;
                    }
                    if (Math.abs(e.pageY - (position.top + position.height)) < delta) {
                        top = position.top + position.height;
                        preview = false;
                    }

                    if (top !== false && ((preview && !Dom.prev(current, canGetFocus, editor.editor)) || (!preview && !Dom.next(current, canGetFocus, editor.editor)))) {
                        line.style.top = top + 'px';
                        show();
                    } else {
                        current = false;
                        hide();
                    }
                }, editor.options.observer.timeout));
        }, editor.options.observer.timeout));
}